import torch


def scale_b_to_a(a: torch.Tensor, b: torch.Tensor):
    """
    Scale tensor b to the same mean and standard deviation as tensor a.

    Args:
        a (torch.Tensor): Target tensor for scaling.
        b (torch.Tensor): Source tensor to be scaled.

    Returns:
        torch.Tensor: Scaled version of tensor b.
    """
    assert not a.requires_grad
    assert a.dim() == b.dim() == 1
    mean, std = a.mean(), a.std()
    sa = a - mean
    sb = b - mean

    if torch.log10(std) < -8:
        std = 1
    scale_a = torch.log10(sa.abs().mean()) - torch.log10(std)
    scale_b = torch.log10(sb.abs().mean()) - torch.log10(std)
    if scale_a > 10 or scale_b > 10:
        std += 1e-6

    sa = sa / std
    sb = sb / std
    return sa, sb



import torch
from torch import Tensor
from torch_geometric.data import Batch

from ..utils.batch import add_graph_attr, add_node_attr, init_batch


def voxel_to_pc(voxel: Tensor, Einc: Tensor) -> Batch:
    """Converts a 3D voxel tensor representation of a particle shower into a hit cloud (pc).
    The conversion is done by finding all non-zero elements in the voxel tensor,
    storing their values and their coordinates in 3D space, and grouping them by shower.

    Parameters
    ----------
    voxel : Tensor
        A 4D tensor representing multiple particle showers in pc_to_voxeld format.
        Dimensions are (num_showers, num_z, num_alpha, num_r).
    Einc : Tensor
        A 1D tensor representing containing the incoming energy of each shower.
        Dimensions are (num_showers,).

    Returns
    -------
    batch: Batch
        A Batch object from the PyTorch Geometric library that contains the point cloud
        representation of the showers, where:
        - batch.batch is a 1D tensor where each element is the ID of the shower the corresponding
        point in the point cloud belongs to.
        - batch.x is a 3D tensor where each row represents a hit in the point cloud,
        with the columns being the value of the voxel and its 3D coordinates (r, alpha, z).
        - batch.y is a 2D tensor where each row represents the incoming energy of the shower
        and the number of non-zero voxels (hits) in it.

    Raises
    ------
    AssertionError
        If there are NaN values in the point cloud tensor x.
    """
    E, showers = Einc.clone(), voxel.clone()
    assert E.dim() == 1
    assert E.dim() < voxel.dim()
    assert (
        E.shape[0] == voxel.shape[0]
    ), "Both inputs must give the same number of events"

    coords = torch.argwhere(
        showers > 0.0
    )  # get indices of non-zero values (shower_id, r, alpha, z)
    vals = showers[tuple(coords.T)]  # get non-zero values
    _, num_hits = torch.unique(
        coords[:, 0], return_counts=True
    )  # get number of non-zero values per shower
    coords = coords[:, 1:]  # remove shower_id from coords
    start_index = torch.zeros(
        num_hits.shape, dtype=torch.int64
    )  # create start_index array
    start_index[1:] = num_hits.cumsum(0)[:-1]  # calculate start_index
    x = torch.hstack((vals[:, None], coords))
    y = torch.stack([E, num_hits], dim=-1)

    # create batch_idx array
    batch_idx = torch.arange(showers.shape[0]).repeat_interleave(num_hits).long()
    batch = init_batch(batch_idx)
    add_node_attr(batch, "x", x)
    add_graph_attr(batch, "y", y)

    return batch



import torch
from torch_scatter import scatter_add

import caloutils
from caloutils.processing import sum_multi_hits

caloutils.init_calorimeter("test")


def test_sum_multi_hits():
    from torch_geometric.data import Batch, Data

    batch = Batch.from_data_list(
        [
            Data(
                x=torch.tensor(
                    [
                        [1, 0, 0, 0],
                    ]
                ),
                y=torch.tensor([[1, 1]]),
                # n_pointsv=torch.tensor(3),
            ),
            Data(
                x=torch.tensor(
                    [
                        [1, 1, 0, 0],
                        [2, 0, 0, 0],
                        [1, 0, 1, 0],
                        [1, 1, 0, 0],
                        [1, 1, 1, 0],
                        [1, 1, 1, 0],
                    ]
                ),
                y=torch.tensor([[1, 1]]),
                # n_pointsv=torch.tensor(4),
            ),
        ]
    )
    batch_new = sum_multi_hits(batch.clone())
    assert (
        batch_new.x
        == torch.tensor(
            [
                [1, 0, 0, 0],
                [2, 0, 0, 0],
                [1, 0, 1, 0],
                [2, 1, 0, 0],
                [2, 1, 1, 0],
            ]
        )
    ).all()

    batchidx = batch.batch
    batchidx_new = batch_new.batch

    n_multihit = batch_new.n_multihit
    new_counts = batch_new.n_pointsv

    hitE = batch.x[:, 0]
    pos = batch.x[:, 1:].long()
    hitE_new = batch_new.x[:, 0]
    pos_new = batch_new.x[:, 1:].long()

    old_counts = torch.unique_consecutive(batchidx, return_counts=True)[1]
    if "n_pointsv" in batch.keys:
        assert (old_counts == batch.n_pointsv).all()
    assert ((old_counts - new_counts) == n_multihit).all()
    assert torch.allclose(
        scatter_add(hitE_new, batchidx_new), scatter_add(hitE, batchidx)
    )
    assert torch.allclose(
        scatter_add(pos_new * hitE_new.unsqueeze(-1), batchidx_new, -2),
        scatter_add(pos * hitE.unsqueeze(-1), batchidx, -2),
    )



from .analyze_layers import analyze_layers
from .calculate_variables import calc_vars
from .energy_ratios import cyratio, sphereratio
from .first_principal_components import first_principal_components
from .response import response



"""Top-level package for caloutils."""

__author__ = """mova"""
__email__ = "moritz.scham@desy.de"
__version__ = '0.0.17'  # fmt: skip

from .calorimeter import calorimeter

init_calorimeter = calorimeter.init_calorimeter
from . import distances, processing, utils, variables



from math import prod

import torch
from torch_scatter import scatter_add

from .. import calorimeter
from .utils import _construct_global_cellevent_idx


def nhits_in_same_cell(batch):
    batch_size = int(batch.batch[-1] + 1)
    dev = batch.x.device
    fulldim = (batch_size, *calorimeter.dims)
    full_event_cell_idx = _construct_global_cellevent_idx(batch_size).to(dev)

    indices = torch.hstack((batch.batch.unsqueeze(1), batch.pos))
    scatter_index = full_event_cell_idx[
        indices[..., 0], indices[..., 1], indices[..., 2], indices[..., 3]
    ]
    occ = scatter_add(
        src=torch.ones(len(batch.batch), dtype=torch.int, device=dev),
        index=scatter_index,
        dim_size=prod(calorimeter.dims) * batch_size,
    ).reshape(*fulldim)
    batch_occ = occ[
        indices[..., 0], indices[..., 1], indices[..., 2], indices[..., 3]
    ]
    return batch_occ



"""
1D Sample Distance Metrics
"""

from typing import Optional

import torch


def energy_distance(
    r: torch.Tensor,
    f: torch.Tensor,
    rw: Optional[torch.Tensor] = None,
    fw: Optional[torch.Tensor] = None,
) -> torch.Tensor:
    """
    Calculate the Energy Distance between two 1D samples.

    The Energy Distance is a statistical measure that quantifies the difference between the empirical distributions
    of two sets of data points.

    Args:
        r (torch.Tensor): Reference sample.
        f (torch.Tensor): Comparison sample.
        rw (Optional[torch.Tensor]): Weights for the reference sample (optional).
        fw (Optional[torch.Tensor]): Weights for the comparison sample (optional).

    Returns:
        torch.Tensor: Calculated Energy Distance between the two samples.
    """
    return _cdf_distance(2, r, f, rw, fw)


def wasserstein_distance(
    r: torch.Tensor,
    f: torch.Tensor,
    rw: Optional[torch.Tensor] = None,
    fw: Optional[torch.Tensor] = None,
) -> torch.Tensor:
    """
    Calculate the Wasserstein Distance (Earth Mover's Distance) between two 1D samples.

    The Wasserstein Distance measures the minimum "cost" of transforming one distribution into another.

    Args:
        r (torch.Tensor): Reference sample.
        f (torch.Tensor): Comparison sample.
        rw (Optional[torch.Tensor]): Weights for the reference sample (optional).
        fw (Optional[torch.Tensor]): Weights for the comparison sample (optional).

    Returns:
        torch.Tensor: Calculated Wasserstein Distance between the two samples.
    """
    return _cdf_distance(1, r, f, rw, fw)


def _cdf_distance(p, u_values, v_values, u_weights=None, v_weights=None):
    """Torch implementation of scipy/scipy/stats /_stats_py.py / _cdf_distance"""
    dev = u_values.device
    u_sorter = torch.argsort(u_values)
    v_sorter = torch.argsort(v_values)

    all_values = torch.cat((u_values, v_values))
    all_values, _ = torch.sort(all_values)

    # Compute the differences between pairs of successive values of u and v.
    deltas = torch.diff(all_values)

    # Get the respective positions of the values of u and v
    #  among the values of both distributions.
    u_cdf_indices = torch.searchsorted(
        u_values[u_sorter], all_values[:-1], right=True
    )
    v_cdf_indices = torch.searchsorted(
        v_values[v_sorter], all_values[:-1], right=True
    )

    # Calculate the CDFs of u and v using their weights, if specified.
    if u_weights is None:
        u_cdf = u_cdf_indices.float() / u_values.numel()
    else:
        u_sorted_cumweights = torch.cat(
            (torch.tensor([0.0]).to(dev), u_weights[u_sorter].cumsum(0))
        )
        u_cdf = u_sorted_cumweights[u_cdf_indices].float() / u_sorted_cumweights[-1]

    if v_weights is None:
        v_cdf = v_cdf_indices.float() / v_values.numel()
    else:
        v_sorted_cumweights = torch.cat(
            (torch.tensor([0.0]).to(dev), v_weights[v_sorter].cumsum(0))
        )
        v_cdf = v_sorted_cumweights[v_cdf_indices].float() / v_sorted_cumweights[-1]

    # Compute the value of the integral based on the CDFs.
    if p == 1:
        return torch.sum(torch.abs(u_cdf - v_cdf) * deltas)
    if p == 2:
        return torch.sqrt(torch.sum(torch.square(u_cdf - v_cdf) * deltas))
    return torch.pow(
        torch.sum(torch.pow(torch.abs(u_cdf - v_cdf), p) * deltas), 1 / p
    )



from torch_geometric.data import Batch

from ..processing import shift_multi_hits, shift_sum_multi_hits, sum_multi_hits
from ..processing.batch_to_Exyz import batch_to_Exyz
from .analyze_layers import analyze_layers
from .energy_ratios import cyratio, sphereratio
from .first_principal_components import first_principal_components
from .response import response


def calc_vars(
    batch: Batch,
    sum_multihits: bool = True,
    shift_multihits: bool = False,
    vars: list = [
        "voxel",
        "sphereratio",
        "cyratio",
        "fpc",
        "showershape",
        "response",
    ],
):
    """
    Calculates specified variables for each event in a batch.

    The function supports calculation of "sphereratio", "cyratio", "fpc", "showershape",
    and "response". Calculated variables are added to the input batch object.

    Parameters
    ----------
    batch : Batch
        A Batch object from the PyTorch Geometric library that contains the point cloud
        representation of the events.
    sum_multihits : bool, optional
        If True, aggregates the energies of hits in the same cell for each event. Defaults to True.
    shift_multihits : bool, optional
        If True, tries to shift hits that occupy the same cell to neihgboring, empty cells. Defaults to False.
    vars : list of str, optional
        List of variables to calculate for the events in the batch. Defaults to all
        supported variables: ["voxel","sphereratio","cyratio","fpc","showershape","response"].

    Returns
    -------
    Batch
        The input Batch object with calculated variables added.
    """
    if sum_multihits and shift_multihits:
        batch = shift_sum_multi_hits(batch)
    elif sum_multihits:
        batch = sum_multi_hits(batch)
    elif shift_multihits:
        batch = shift_multi_hits(batch)

    batch = batch_to_Exyz(batch)
    if "sphereratio" in vars:
        batch["sphereratio"] = sphereratio(batch)
    if "cyratio" in vars:
        batch["cyratio"] = cyratio(batch)
    if "fpc" in vars:
        batch["fpc"] = first_principal_components(batch)
    if "showershape" in vars:
        batch["showershape"] = analyze_layers(batch)
    if "response" in vars:
        batch["response"] = response(batch)
    return batch



from collections import defaultdict

import torch
import torch_scatter
from torch import Tensor
from torch_geometric.data import Batch
from torch_geometric.nn import pool

from .. import calorimeter


def scatter_sort(x: Tensor, index: Tensor, dim: int = -1, descending: bool = False):
    """
    Sorts the elements of `x` that share the same `index`, returns sorted `x` and `index` needed.

    Parameters
    ----------
    x : torch.Tensor
        The input tensor to sort.
    index : torch.Tensor
        The index tensor to sort according to `x`.
    dim : int, optional
        The dimension along which to sort. Defaults to the last dimension.
    descending : bool, optional
        Lowest value first. Defaults to true.

    Returns
    -------
    x : torch.Tensor
        The sorted input tensor.
    index : torch.Tensor
        The index tensor sorted according to `x`.
    """
    x, x_perm = torch.sort(x, dim=dim, descending=descending)
    index = index.take_along_dim(x_perm, dim=dim)
    index, index_perm = torch.sort(index, dim=dim, stable=True)
    x = x.take_along_dim(index_perm, dim=dim)
    return x, x_perm.take_along_dim(index_perm, dim=dim)


def ptr_from_batchidx(batch_idx: Tensor):
    """
    Constructs a pointer tensor from the given batch index tensor. The pointer tensor allows to address
    individual graphs within a batch of graphs.

    The constructed pointer tensor corresponds to the cumulative sum of the number of nodes in each individual graph.
    This allows to index specific graphs within the batch.

    Parameters
    ----------
    batch_idx : torch.Tensor
        1D tensor containing the batch indices for each node in the batch.
        Each value in batch_idx indicates which graph the node belongs to.
        For example, if batch_idx[i] = j, then the ith node belongs to the jth graph in the batch.

    Returns
    -------
    ptr : torch.Tensor
        Pointer tensor constructed from the batch_idx. This tensor can be used to index individual graphs in the batch.
        For example, if ptr[k] = m, then the first node of the kth graph in the batch is the mth node in the node feature matrix.

    Examples
    --------
    >>> batch_idx = torch.tensor([0, 0, 1, 1, 1, 2])
    >>> ptr_from_batchidx(batch_idx)
    tensor([0, 2, 5, 6])
    """
    dev = batch_idx.device
    return torch.concatenate(
        (
            torch.tensor(0).long().to(dev).unsqueeze(0),
            (batch_idx.diff()).nonzero().reshape(-1) + 1,
            torch.tensor(len(batch_idx)).long().to(dev).unsqueeze(0),
        )
    )


def fix_slice_dict_nodeattr(batch: Batch, attrname: str) -> Batch:
    """
    Modifies the "_slice_dict" attribute of the Batch object to include the number of elements of a given attribute
    per graph in the batch.

    Parameters
    ----------
    batch : Batch
        The batch of graphs.
    attrname : str
        The attribute of the batch for which the slice dictionary is to be updated.

    Returns
    -------
    batch : Batch
        The modified batch object with updated "_slice_dict".

    Note
    ----
    The "_slice_dict" attribute of the batch object is a dictionary which stores the number of elements
    of various attributes per graph in the batch. This function updates this dictionary for a given attribute.
    """
    if not hasattr(batch, "_slice_dict"):
        batch._slice_dict = defaultdict(dict)
    attr = batch[attrname]
    batch_idxs = batch.batch
    device = attr.device
    out = torch_scatter.scatter_add(
        torch.ones(len(attr), dtype=torch.long, device=device), batch_idxs, dim=0
    )
    out = out.cumsum(dim=0)
    batch._slice_dict[attrname] = torch.cat(
        [torch.zeros(1, dtype=torch.long, device=device), out], dim=0
    )
    return batch


def global_std_pool(x: Tensor, batchidx: Tensor) -> Tensor:
    """
    Computes the global standard deviation of node features in a batch of graphs.

    Parameters
    ----------
    x : Tensor
        The node features of the batch.
    batchidx : Tensor
        The batch indices for each node in the batch.

    Returns
    -------
    Tensor
        A tensor containing the global standard deviation of node features for each graph in the batch.
    """
    return torch.sqrt(global_var_pool(x, batchidx))


def global_var_pool(x: Tensor, batchidx: Tensor) -> Tensor:
    """
    Computes the global variance of node features in a batch of graphs.

    Parameters
    ----------
    x : Tensor
        The node features of the batch.
    batchidx : Tensor
        The batch indices for each node in the batch.

    Returns
    -------
    Tensor
        A tensor containing the global variance of node features for each graph in the batch.
    """
    means = pool.global_mean_pool(x, batchidx)
    deltas = torch.pow(means[batchidx] - x, 2)
    return pool.global_mean_pool(deltas, batchidx)


def _construct_global_cellevent_idx(batch_size: int) -> torch.Tensor:
    dims = calorimeter.dims
    full_event_cell_idx = torch.arange(
        batch_size * dims[0] * dims[1] * dims[2]
    ).reshape(batch_size, *dims)
    return full_event_cell_idx



"""
ecdf: Calculate an approximation to the cumulative distribution function (CDF).
ecdf_distance: Calculate a distance metric between two 1D samples using ecdf.
calc_ecdf_dist: Convenience function to run ecdf_distance over multiple dimensions and convert the result to a numpy array.
"""

from typing import Optional

import numpy as np
import torch


def ecdf(
    arr: torch.Tensor, weights: Optional[torch.Tensor] = None
) -> tuple[torch.Tensor, Optional[torch.Tensor]]:
    """
    Calculate an approximation to the cumulative distribution function (CDF) using empirical distribution.


    Args:
        arr (torch.Tensor): Input tensor containing data points.
        weights (Optional[torch.Tensor]): Weights for each data point (optional).

    Returns:
        tuple[torch.Tensor, Optional[torch.Tensor]]: Tuple containing the calculated CDF values and weights (if provided).
    """
    val, sortidx = arr.sort()
    cdf = val.cumsum(-1)
    cdf = cdf.clone() / cdf[-1].clone()
    if weights is None:
        return cdf, None
    else:
        return cdf, weights[sortidx]


def ecdf_distance(
    r: torch.Tensor,
    f: torch.Tensor,
    rw: Optional[torch.Tensor] = None,
    fw: Optional[torch.Tensor] = None,
) -> torch.Tensor:
    """
    Calculate a distance metric between two 1D samples using the ecdf function.

    Args:
        r (torch.Tensor): Reference sample.
        f (torch.Tensor): Comparison sample.
        rw (Optional[torch.Tensor]): Weights for the reference sample (optional).
        fw (Optional[torch.Tensor]): Weights for the comparison sample (optional).

    Returns:
        torch.Tensor: Calculated distance metric between the two samples.
    """
    real_cdf, w_real = ecdf(r, rw)
    cdf_fake, w_fake = ecdf(f, fw)
    dist = cdf_fake - real_cdf
    if rw is not None:
        ww = w_fake * w_real
        ww /= ww.sum()
        dist *= ww
    return dist.abs().mean(0)


def calc_ecdf_dist(
    r: torch.Tensor,
    f: torch.Tensor,
    rw: Optional[torch.Tensor] = None,
    fw: Optional[torch.Tensor] = None,
) -> np.ndarray:
    """
    Calculate distance metrics between two samples over multiple dimensions using ecdf_distance and convert to a numpy array.

    Args:
        r (torch.Tensor): Reference sample.
        f (torch.Tensor): Comparison sample.
        rw (Optional[torch.Tensor]): Weights for the reference sample (optional).
        fw (Optional[torch.Tensor]): Weights for the comparison sample (optional).

    Returns:
        np.ndarray: Array of distance metrics calculated for each dimension.
    """
    assert r.shape == f.shape
    assert (rw is None) == (fw is None)
    dists = []
    for iftx in range(r.shape[-1]):
        dists.append(ecdf_distance(r[..., iftx], f[..., iftx], rw, fw))
    return torch.stack(dists, dim=0).cpu().numpy()



import torch
from torch import Tensor
from torch_geometric.data import Batch
from torch_geometric.nn.pool import global_max_pool

from .. import calorimeter


def analyze_layers(batch: Batch) -> dict[str, Tensor]:
    """
    Analyzes the layers of a point cloud representation of particle showers, computing
    the peak layer, the turn-on layer, and the ratio of distances from turn-off layer to
    the peak layer and from the peak layer to the turn-on layer.

    The function performs the following main steps:
    1. Computes a histogram of the z coordinates (layers) of each point in the point cloud.
    2. Identifies the layer with the maximum number of points (peak layer) for each shower.
    3. Identifies the first (turn-on) and last (turn-off) layer that contains at least half
    the maximum number of points in a layer for each shower.
    4. Computes the ratio of the distance from the turn-off layer to the peak layer and
    the distance from the peak layer to the turn-on layer.

    Parameters
    ----------
    batch : Batch
        A Batch object from the PyTorch Geometric library that contains the point cloud
        representation of the showers.

    Returns
    -------
    dict
        A dictionary with keys 'peak_layer', 'psr', and 'turnon_layer', where:
            - 'peak_layer' corresponds to the tensor of peak layers for each shower in the batch.
            - 'psr' corresponds to the tensor of ratios for each shower in the batch.
            - 'turnon_layer' corresponds to the tensor of turn-on layers for each shower in the batch.

    Raises
    ------
    AssertionError
        If the sum of the elements in each row of the histogram does not equal
        the difference of consecutive elements in the batch pointers.

    """
    batchidx = batch.batch
    n_events = int(batchidx[-1] + 1)
    device = batch.x.device
    z = batch.xyz[..., -1]
    # compute the histogramm with dim (events,layers)
    z_ev = (z.long() + batchidx * calorimeter.num_z).sort().values

    zvals, zcounts = z_ev.unique_consecutive(return_counts=True)

    layer_idx = zvals % calorimeter.num_z
    ev_idx = (zvals - layer_idx) // calorimeter.num_z

    hist = torch.zeros(n_events, calorimeter.num_z).long().to(device)
    hist[ev_idx, layer_idx] = zcounts

    assert (hist.sum(1) == batch.ptr[1:] - batch.ptr[:-1]).all()

    max_hits_per_event, peak_layer_per_event = hist.max(1)

    occupied_layers = (
        hist > max_hits_per_event.reshape(-1, 1).repeat(1, calorimeter.num_z) / 2
    )
    eventidx, occ_layer = occupied_layers.nonzero().T
    # find the min and max layer from the bool matrix ath the same time
    minmax = global_max_pool(torch.stack([occ_layer, -occ_layer]).T, eventidx)
    # reverse the -1 trick
    turnoff_layer, turnon_layer = (minmax * torch.tensor([1, -1]).to(device)).T

    # distance to the peak
    turnoff_dist = turnoff_layer - peak_layer_per_event
    turnon_dist = peak_layer_per_event - turnon_layer
    psr = (turnoff_dist + 1) / (turnon_dist + 1)

    return {
        "peak_layer": peak_layer_per_event.float(),
        "psr": psr.float(),
        "turnon_layer": turnon_layer.float(),
    }



import torch
from torch_geometric.data import Batch
from torch_geometric.nn import global_mean_pool

from ..processing.utils import global_std_pool


def first_principal_components(batch: Batch) -> dict[str, torch.Tensor]:
    """
    Calculates the first principal component (PC) from a point cloud (PC).

    Parameters
    ----------
    batch : Batch
        A Batch object from the PyTorch Geometric library that contains the point cloud
        representation of the events. Batch.xyz contains the 3D coordinates of hits.
        Batch.batch contains the indices that map which hit belongs to which event.

    Returns
    -------
    fpc : dict
        Dictionary containing the components of the first principal component and the corresponding
        eigenvalue for each event. The keys of the dictionary are 'x', 'y', 'z', and 'eval'.
    """
    batchidx = batch.batch
    xyz = batch.xyz.double()
    means = global_mean_pool(xyz, batchidx)
    stds = global_std_pool(xyz, batchidx)
    deltas = (xyz - means[batchidx]) / (stds[batchidx] + 1e-8)
    cov = torch.stack(
        [
            torch.stack(
                [
                    global_mean_pool(deltas[:, i] * deltas[:, j], batchidx)
                    # global_add_pool(deltas[:, i] * deltas[:, j], batchidx)
                    # / (batch.ptr[1:]-batch.ptr[:-1] - 1)  # normalize
                    for i in range(3)
                ]
            )
            for j in range(3)
        ]
    ).transpose(0, -1)

    e_vals, e_vec = torch.linalg.eigh(cov)

    # largest_ev = e_val.argmax(-1).reshape(-1, 1, 1)
    # first_pc = e_vec.take_along_dim(largest_ev, -1)
    # (first_pc==e_vec[:,:,[2]]).all() why?? are they sorted?
    # https://pytorch.org/docs/stable/generated/torch.linalg.eigh.html
    # e_vec are sorted, get last colomn
    first_pc = e_vec[:, :, -1]

    # TEST
    # untrfs = deltas[batchidx==0]
    # first pc in the the first component:
    # fct = (first_pc[0].reshape(1, 3) @ untrfs.T).reshape(-1, 1) * first_pc[
    #     0
    # ].reshape(1, 3)
    # assert ((untrfs-fct).std(0)<(untrfs).std(0)).all()
    return dict(zip(["x", "y", "z"], first_pc.T.float().abs())) | {
        "eval": e_vals[:, -1].float()
    }



from .batch_to_Exyz import batch_to_Exyz
from .nhits_in_same_cell import nhits_in_same_cell
from .pc_from_voxel import voxel_to_pc
from .pc_to_voxel import pc_to_voxel
from .shift_multi_hits import shift_multi_hits
from .sum_multi_hits import shift_sum_multi_hits, sum_multi_hits



from typing import Optional

import torch
from torch_geometric.data import Batch

from .. import calorimeter
from .utils import scatter_sort


def shift_multi_hits(
    batch: Batch,
    globalidx: Optional[torch.Tensor] = None,
    return_globalidx: bool = False,
):
    """
    Shifts hits that belong to the same cell, to neighboring, empty cells, if available.
    Leaves the highest energy hit in the cell and prioritizes moving the 2nd most energetic (then 3rd most ...) hit.
    The function modifies the batch in place, updating batch.x, batch.batch, and batch.ptr.

    Parameters
    ----------
    batch : Batch
        A Batch object from the PyTorch Geometric library that contains the point cloud
        representation of the events. Batch.x contains the hit energy and 3D coordinates of hits.
        Batch.batch contains the indices that map which hit belongs to which shower.

    forbid_dublicates : bool, optional
        If True, asserts that there were no duplicate hits in the original data. Defaults to True.

    Returns
    -------
    batch : Batch
        The modified Batch object where duplicate hits have been summed.
    """
    batchidx: torch.Tensor = batch.batch
    dev = batch.x.device

    if globalidx is None:
        globalidx = calorimeter.globalidx_from_pos(batch.x[:, 1:].long(), batchidx)

    # sort, so we can check globalidx for repetitions to check for double cells
    globalidx, index_perm = scatter_sort(globalidx, batchidx)
    batch.x = batch.x[index_perm]

    # The idea of this algorithm is to look for repetitions
    # in the (sorted) globalidx to detect repetitions.

    global_uniques, unique_cells_idx, counts = torch.unique(
        globalidx, return_inverse=True, return_counts=True
    )

    # Because we want to keep the highest energy hit in the same
    # cell we now sort the hits in the same cell by the energy
    # For this, we need to know which hits share a cell
    has_multihits = torch.isin(globalidx, global_uniques[counts > 1])
    # Here we want to sort energies of the same cell/event,
    # so we use the globalidx instead of the usual batchidx
    _, perm = scatter_sort(
        batch.x[..., 0][has_multihits], globalidx[has_multihits], descending=True
    )
    batch.x[has_multihits] = batch.x[has_multihits][perm]
    globalidx[has_multihits] = globalidx[has_multihits][perm]
    assert (batchidx[has_multihits] == batchidx[has_multihits][perm]).all()

    # select the second hit in each cell for moving around
    mhit_idxs = (
        torch.cat([torch.tensor([0]).to(dev), 1 - unique_cells_idx.diff()])
        .nonzero()
        .squeeze()
    )
    shift_options = len(calorimeter.dims) * 2
    # start value for each shift
    shift_state = torch.randint_like(mhit_idxs, 1, shift_options)

    # switch for testing this function
    if calorimeter._caloname == "test":
        shift_state = torch.ones_like(shift_state)
    idxs_to_overwrite = []
    new_x_list = []
    new_global_list = []

    for _ in range(shift_options):
        if len(mhit_idxs) == 0:
            break

        batchidx_to_shift = batchidx[mhit_idxs]
        new_x = _shift_pos(batch.x[mhit_idxs], shift_state)
        # get the new globals index for the shifted hits
        new_global = calorimeter.globalidx_from_pos(new_x[:, 1:], batchidx_to_shift)

        # Check if the shift is valid
        # A) Check if this is the first hit to be shifted to this cell
        # we now need and index that filters mhit_idxs
        # and only lets the first though.
        # (4,3,4,1,3)   -> (1,1,0,1,0) or (0,1,3)
        # we can check if the index  of new_global.unique is increaseing
        # but for this we need to sort the indexes again
        new_global, perm = new_global.sort()
        mhit_idxs = mhit_idxs[perm]
        new_x = new_x[perm]

        # now we can check of the unique rev index is increasing
        # which tells us of the cell is being accessed for the first time

        first_new_hit = torch.cat(
            (
                torch.tensor([1]).to(dev),
                new_global.unique(return_inverse=True)[1].diff(),
            )
        ).bool()

        # B)
        # check if the new position is valid iff it's not already occupied
        new_pos_is_free = ~torch.isin(
            new_global, torch.cat([globalidx, *new_global_list])
        )

        # combine the two
        valide_shift_index = new_pos_is_free & first_new_hit

        # apppend the shifted hits to the list
        if valide_shift_index.sum():
            idxs_to_overwrite.append(mhit_idxs[valide_shift_index])
            new_x_list.append(new_x[valide_shift_index])
            new_global_list.append(new_global[valide_shift_index])
        # rerun loop with the remaining multihits
        mhit_idxs = mhit_idxs[~valide_shift_index]
        shift_state = (shift_state[~valide_shift_index] + 1) % shift_options

    if len(idxs_to_overwrite):
        # overwrite old position and global index
        stacked_idxs_to_overwrite = torch.cat(idxs_to_overwrite)
        globalidx[stacked_idxs_to_overwrite] = torch.cat(new_global_list)
        batch.x[stacked_idxs_to_overwrite] = torch.cat(new_x_list)

    if return_globalidx:
        return batch, globalidx
    else:
        return batch


def _shift_pos(x, shift_state):
    pos = x[:, 1:].long()
    dev = pos.device
    directon = (shift_state % 2) * 2 - 1
    dim = shift_state // 2
    pos[torch.arange(len(pos)), dim] += directon

    # rotate alpha
    num_alpha = calorimeter.num_alpha
    alphas = pos[:, 1]
    alphas[alphas > num_alpha - 1] -= num_alpha
    alphas[alphas < 0] -= num_alpha
    pos[:, 1] = alphas

    # clamp r and z
    pos = torch.clamp(
        pos,
        torch.tensor([0, 0, 0]).to(dev),
        torch.tensor(calorimeter.dims).to(dev) - 1,
    )
    x[:, 1:] = pos
    return x



import torch
from torch_geometric.data import Batch
from torch_geometric.nn import global_add_pool


def response(batch: Batch) -> torch.Tensor:
    """
    Computes the energy response for each event in a batch.
    The energy response is calculated as the ratio of the total detected energy
    (sum of energy from all hits in an event) to the true energy of the event.

    Parameters
    ----------
    batch : Batch
        A Batch object from the PyTorch Geometric library that contains the point cloud
        representation of the events.

    Returns
    -------
    torch.Tensor
        A tensor of the energy response for each event in the batch.
    """

    batchidx = batch.batch
    Ehit = batch.x[:, 0]
    Esum = global_add_pool(Ehit, batchidx)
    return Esum / batch.y[:, 0]



import torch
from torch_geometric.data import Batch, Data

from caloutils.utils.batch import (
    add_graph_attr,
    add_node_attr,
    from_batch_list,
    init_batch,
    set_edge_attr,
    set_edges,
)

device = torch.device("cpu")


def test_batch_add_node_attr():
    batch_size = 3
    node_range = (1, 10)
    nodes_v = torch.randint(*node_range, (batch_size,)).to(device)
    x_list = [torch.rand(n, 3).to(device) for n in nodes_v]

    batch_list = [Data(x=x) for x in x_list]
    batch_truth = Batch.from_data_list(batch_list)

    batch_idx = batch_truth.batch

    batch = init_batch(batch_idx)

    add_node_attr(batch, "x", torch.vstack(x_list))

    compare(batch_truth, batch)


def test_batch_set_edges():
    batch_size = 4
    node_range = (2, 5)
    nodes_v = torch.randint(*node_range, (batch_size,)).to(device)
    edges_per_graph = torch.cat(
        [torch.randint(1, num_nodes, size=(1,)).to(device) for num_nodes in nodes_v]
    ).to(device)
    x_list = [torch.rand(num_nodes, 3).to(device) for num_nodes in nodes_v]
    edges_list = [
        torch.vstack(
            [
                torch.randint(0, num_nodes, size=(num_edges,)),
                torch.randint(0, num_nodes, size=(num_edges,)),
            ]
        ).to(device)
        for num_nodes, num_edges in zip(nodes_v, edges_per_graph)
    ]

    batch_list = [
        Data(x=x, edge_index=edges) for x, edges in zip(x_list, edges_list)
    ]
    batch_truth = Batch.from_data_list(batch_list)

    batch_idx = batch_truth.batch
    batch = init_batch(batch_idx)
    add_node_attr(batch, "x", torch.vstack(x_list))

    batchidx_per_edge = torch.cat(
        [
            torch.ones(num_edges).long().to(device) * igraph
            for igraph, num_edges in enumerate(edges_per_graph)
        ]
    )
    set_edges(batch, torch.hstack(edges_list), batchidx_per_edge)
    compare(batch_truth, batch)


def test_batch_set_edge_attr():
    batch_size = 4
    node_range = (2, 5)
    nodes_v = torch.randint(*node_range, (batch_size,)).to(device)
    edges_per_graph = torch.cat(
        [torch.randint(1, num_nodes, size=(1,)).to(device) for num_nodes in nodes_v]
    )
    x_list = [torch.rand(num_nodes, 3).to(device) for num_nodes in nodes_v]
    edges_list = [
        torch.vstack(
            [
                torch.randint(0, num_nodes, size=(num_edges,)),
                torch.randint(0, num_nodes, size=(num_edges,)),
            ]
        ).to(device)
        for num_nodes, num_edges in zip(nodes_v, edges_per_graph)
    ]
    edge_attr_list = [
        torch.rand(num_edges).to(device) for num_edges in edges_per_graph
    ]

    batch_list = [
        Data(x=x, edge_index=edges, edge_attr=ea)
        for x, edges, ea in zip(x_list, edges_list, edge_attr_list)
    ]
    batch_truth = Batch.from_data_list(batch_list)

    batch_idx = batch_truth.batch
    batch = init_batch(batch_idx)
    add_node_attr(batch, "x", torch.vstack(x_list))

    batchidx_per_edge = torch.cat(
        [
            torch.ones(num_edges).to(device).long() * igraph
            for igraph, num_edges in enumerate(edges_per_graph)
        ]
    )
    set_edges(batch, torch.hstack(edges_list), batchidx_per_edge)
    set_edge_attr(batch, torch.hstack(edge_attr_list), batchidx_per_edge)
    compare(batch_truth, batch)


def test_batch_add_graph_attr():
    batch_size = 3
    node_range = (1, 10)
    nodes_v = torch.randint(*node_range, (batch_size,)).to(device)
    x_list = [torch.rand(n, 3).to(device) for n in nodes_v]
    graph_attr_list = torch.rand(batch_size).to(device)

    batch_list = [Data(x=x, ga=ga) for x, ga in zip(x_list, graph_attr_list)]
    batch_truth = Batch.from_data_list(batch_list)

    batch_idx = batch_truth.batch

    batch = init_batch(batch_idx)

    add_node_attr(batch, "x", torch.vstack(x_list))
    add_graph_attr(batch, "ga", graph_attr_list)
    compare(batch_truth, batch)


def test_from_batch_list():
    batch_size = 9
    node_range = (2, 5)
    nodes_v = torch.randint(*node_range, (batch_size,)).to(device)
    edges_per_graph = torch.cat(
        [torch.randint(1, num_nodes, size=(1,)).to(device) for num_nodes in nodes_v]
    )
    x_list = [torch.rand(num_nodes, 3).to(device) for num_nodes in nodes_v]
    edges_list = [
        torch.vstack(
            [
                torch.randint(0, num_nodes, size=(num_edges,)),
                torch.randint(0, num_nodes, size=(num_edges,)),
            ]
        ).to(device)
        for num_nodes, num_edges in zip(nodes_v, edges_per_graph)
    ]
    edge_attr_list = [
        torch.rand(num_edges).to(device) for num_edges in edges_per_graph
    ]
    graph_attr_list = torch.rand(batch_size).to(device)

    batch_list = [
        Data(x=x, edge_index=edges, edge_attr=ea, ga=ga)
        for x, edges, ea, ga in zip(
            x_list, edges_list, edge_attr_list, graph_attr_list
        )
    ]
    batch_truth = Batch.from_data_list(batch_list)
    batch = from_batch_list(
        Batch.from_data_list(batch_list[:3]), Batch.from_data_list(batch_list[3:])
    )

    compare(batch_truth, batch)


def compare(ba: Batch, bb: Batch):
    if set(ba.keys) != set(bb.keys):
        raise Exception()
    for k in ba.keys:
        rec_comp(ba[k], bb[k])
        rec_comp(ba._slice_dict[k], bb._slice_dict[k])
        rec_comp(ba._inc_dict[k], bb._inc_dict[k])


def rec_comp(a, b):
    if not type(a) == type(b):
        raise Exception()
    if isinstance(a, dict):
        if not set(a.keys()) == set(b.keys()):
            raise Exception()
        for k in a:
            rec_comp(a[k], b[k])
    if isinstance(a, torch.Tensor):
        if not (a == b).all():
            raise Exception()


test_batch_set_edges()



from .ecdf import calc_ecdf_dist, ecdf, ecdf_distance
from .hist import calc_hist_dist, hist_distance, histcdf
from .scaled_w1 import calc_sw1_dist, scaled_w1_distance
from .wasserstein import energy_distance, wasserstein_distance



import torch
from torch_scatter import scatter_add

import caloutils
from caloutils.processing import shift_multi_hits

caloutils.init_calorimeter("test")


def testshift_multi_hits():
    from torch_geometric.data import Batch, Data

    batch = Batch.from_data_list(
        [
            Data(x=torch.tensor([[7, 0, 0, 0]])),
            Data(
                x=torch.tensor(
                    [
                        [1, 0, 0, 0],
                        [1, 0, 0, 0],
                        [3, 0, 0, 0],
                        [1, 0, 0, 0],
                    ]
                )
            ),
            Data(x=torch.tensor([[1, 1, 1, 0], [2, 1, 1, 0]])),
        ]
    )
    batch_new = shift_multi_hits(batch.clone())

    assert (
        batch_new.x
        == torch.tensor(
            [
                [7, 0, 0, 0],
                [3, 0, 0, 0],
                [1, 1, 0, 0],
                [1, 0, 1, 0],
                [1, 0, 0, 0],
                [2, 1, 1, 0],
                [1, 1, 0, 0],
            ]
        )
    ).all()

    batchidx = batch.batch

    old_counts = torch.unique_consecutive(batchidx, return_counts=True)[1]
    if "n_pointsv" in batch.keys:
        assert (old_counts == batch.n_pointsv).all()
        assert ((old_counts - batch_new.n_pointsv) == batch_new.n_multihit).all()
    assert torch.allclose(
        scatter_add(batch_new.x[:, 0], batch_new.batch),
        scatter_add(batch.x[:, 0], batchidx),
    )



from functools import partial

import torch

from caloutils.distances import (
    calc_ecdf_dist,
    calc_hist_dist,
    calc_sw1_dist,
    ecdf_distance,
    energy_distance,
    hist_distance,
    scaled_w1_distance,
    wasserstein_distance,
)


def test_distances():
    x = torch.tensor([1, 0, 0, 0]).float()
    y = torch.tensor([0, 1, 2, 0]).float()
    w = torch.tensor([1, 2, 1, 1]).float()
    bins = torch.arange(0, 4, step=0.5)

    for dist in [
        wasserstein_distance,
        energy_distance,
        ecdf_distance,
        partial(hist_distance, bins=bins),
        scaled_w1_distance,
    ]:
        assert dist(x, x, rw=w, fw=w) == 0
        assert dist(x, y, rw=w, fw=w) > 0

    x = x.unsqueeze(1)
    y = y.unsqueeze(1)

    for dist in [
        calc_ecdf_dist,
        calc_sw1_dist,
        partial(calc_hist_dist, bins=bins.unsqueeze(1)),
    ]:
        assert all((dist(x, x, rw=w, fw=w) == 0))
        assert all(dist(x, y, rw=w, fw=w) >= 0)



from typing import Optional

import numpy as np
import torch

from ..utils import scale_b_to_a
from .wasserstein import wasserstein_distance


def scaled_w1_distance(
    r: torch.Tensor,
    f: torch.Tensor,
    rw: Optional[torch.Tensor] = None,
    fw: Optional[torch.Tensor] = None,
) -> torch.Tensor:
    """
    Calculate the scaled Wasserstein-1 Distance between two 1D samples.

    The scaled Wasserstein-1 Distance measures the difference between the distributions of two samples
    while accounting for differences in scale and location.

    Args:
        r (torch.Tensor): Reference sample.
        f (torch.Tensor): Comparison sample.
        rw (Optional[torch.Tensor]): Weights for the reference sample (optional).
        fw (Optional[torch.Tensor]): Weights for the comparison sample (optional).

    Returns:
        torch.Tensor: Calculated scaled Wasserstein-1 Distance between the two samples.
    """
    assert r.dim() == f.dim() == 1
    rs, fs = scale_b_to_a(r, f)
    return wasserstein_distance(rs, fs, rw, fw)


def calc_sw1_dist(
    r: torch.Tensor,
    f: torch.Tensor,
    rw: Optional[torch.Tensor] = None,
    fw: Optional[torch.Tensor] = None,
) -> np.ndarray:
    """
    Calculate scaled Wasserstein-1 distance metrics between two samples over multiple dimensions.

    Args:
        r (torch.Tensor): Reference sample.
        f (torch.Tensor): Comparison sample.
        rw (Optional[torch.Tensor]): Weights for the reference sample (optional).
        fw (Optional[torch.Tensor]): Weights for the comparison sample (optional).

    Returns:
        np.ndarray: Array of scaled Wasserstein-1 distance metrics calculated for each dimension.
    """
    assert r.shape == f.shape
    assert (rw is None) == (fw is None)
    dists = []
    for iftx in range(r.shape[-1]):
        cdfdist = scaled_w1_distance(r[..., iftx], f[..., iftx], rw, fw)
        dists.append(cdfdist)
    return torch.stack(dists, axis=0).cpu().numpy()



# # %%
# import numpy as np
# import torch
# from fgsim.config import compute_conf, conf
# from torch_geometric.data import Batch
# from tqdm import tqdm

# compute_conf(conf, {"dataset_name": "calochallange", "command": "test"})
# step = 1_000
# conf.loader.scaling_fit_size //= 10
# conf.loader.batch_size = step

# from fgsim.io.dequantscaler import (  # noqa: E402
#     FunctionTransformer,
#     dequant,
#     requant,
#     scipy,
# )
# from fgsim.loaders.calochallange.objcol import (  # noqa: E402
#     contruct_graph_from_row,
#     file_manager,
#     read_chunks,
#     scaler,
# )
# from fgsim.loaders.calochallange.pc_to_voxel import (  # noqa: E402
#     dims,
#     pc_to_voxel,
#     sum_multi_hits,
# )

# # scaler.fit(True)
# # print("fit done")

# for iftx in range(1, 4):
#     sel_tf = scaler.transfs_x[iftx]
#     sel_tf.steps[0][1].func = dequant
#     sel_tf.steps[0][1].inverse_func = requant

#     sel_tf.steps[2][1].func = scipy.special.logit
#     sel_tf.steps[2][1].inverse_func = scipy.special.expit


# fpos = np.arange(0, 100_000, step)
# for i in tqdm(fpos):
#     pcs = read_chunks([(file_manager.files[0], i, i + step)])
#     gl = []
#     for pc in pcs:
#         graph = contruct_graph_from_row(pc)
#         graph.n_pointsv = (
#             graph.y[..., conf.loader.y_features.index("num_particles")]
#             .int()
#             .reshape(-1)
#         )

#         gl.append(graph)

#     batch = Batch.from_data_list(gl)
#     batch_untf = batch.clone()
#     batch.y = scaler.transform(batch.y, "y")

#     ## Manual scaling
#     # maxscaler = 1
#     maxscaler = 5
#     x = batch.x.clone()
#     # forward
#     for iftx in range(4):
#         ix = x[:, [iftx]].double().numpy()
#         sel_tf = scaler.transfs_x[iftx]
#         tfsteps = [e[1] for e in sel_tf.steps][: maxscaler + 1]

#         for itf in tfsteps:
#             x_tf = itf.transform(ix)
#             x_tf_inv = itf.inverse_transform(x_tf)
#             assert np.allclose(x_tf_inv, ix)
#             ix = x_tf
#         x[:, [iftx]] = torch.tensor(ix).float()

#     x_forward = x.clone()

#     # backward
#     for iftx in range(4):
#         ix = x[:, [iftx]].double().numpy()
#         sel_tf = scaler.transfs_x[iftx]
#         tfsteps = [e[1] for e in sel_tf.steps][: maxscaler + 1]

#         for itf in tfsteps[::-1]:
#             x_tf = itf.inverse_transform(ix)
#             x_tf_inv = itf.transform(x_tf)

#             if isinstance(itf, FunctionTransformer) and itf.func == dequant:
#                 assert np.allclose(x_tf_inv.astype("int"), ix.astype("int"))
#             else:
#                 assert np.allclose(x_tf_inv, ix)
#             ix = x_tf
#         x[:, [iftx]] = torch.tensor(ix).float()

#     assert torch.allclose(x, batch_untf.x)

#     # batch.x = x
#     batch.y = scaler.inverse_transform(batch.y, "y")

#     ## Scaler scaling
#     batch.x = scaler.transform(batch_untf.x.clone(), "x")
#     batch_scaled = batch.clone()
#     _inv = scaler.inverse_transform(batch.x, "x")
#     assert torch.allclose(_inv, x)
#     assert torch.allclose(_inv, batch_untf.x)
#     batch.x = _inv

#     # alphapos = conf.loader.x_features.index("alpha")
#     # batch.x[..., alphapos] = rotate_alpha(
#     #     batch.x[..., alphapos].clone(), batch.batch, True
#     # )

#     delta = (batch.x - batch_untf.x).abs()
#     idx = torch.where(delta == delta.max())

#     if not torch.allclose(batch.x, batch_untf.x, rtol=1e-04):
#         sel_tf = scaler.transfs_x[idx[1]]
#         print(batch.x[idx], batch_untf.x[idx], idx[1])
#         print(
#             batch_scaled.x[idx],
#             sel_tf.transform(batch_untf.x[:, idx[1]]).squeeze()[idx[0]],
#             idx[1],
#         )

#         tfsteps = [e[1] for e in sel_tf.steps]
#         forward_res_step = [batch_untf.x[idx].numpy().reshape(-1, 1)]
#         for _tf in tfsteps:
#             forward_res_step.append(_tf.transform(forward_res_step[-1].copy()))

#         backward_res_step = [batch_scaled.x[idx].numpy().reshape(-1, 1)]
#         print(f"tranformed {backward_res_step[-1]}, scaled {forward_res_step[-1]}")
#         for itf, _tf in enumerate(tfsteps[::-1]):
#             backward_res_step.append(
#                 _tf.inverse_transform(backward_res_step[-1].copy())
#             )
#             a = backward_res_step[-1].squeeze()
#             b = forward_res_step[-2 - itf].squeeze()
#             print(a, b, "out of", _tf)

#         raise Exception()

#     assert torch.allclose(batch.x, batch_untf.x, rtol=1e-04)

#     ## cc postprocess step by step

#     # alphashift
#     # alphas = batch.x[..., alphapos].clone()
#     # shift = torch.randint(0, num_alpha,
#     #  (batch.batch[-1] + 1,)).to(alphas.device)[
#     #     batch.batch
#     # ]
#     # alphas = alphas.clone() + shift.float()
#     # alphas[alphas > num_alpha - 1] -= num_alpha
#     # batch.x[..., alphapos] = alphas

#     batch = sum_multi_hits(batch, True)
#     assert torch.allclose(batch.x, batch_untf.x)

#     # batch = batch_to_Exyz(batch)
#     # metrics: list[str] = conf.training.val.metrics
#     # if "sphereratio" in metrics:
#     #     batch["sphereratio"] = sphereratio(batch)
#     # if "fpc" in metrics:
#     #     batch["fpc"] = fpc_from_batch(batch)
#     # if "showershape" in metrics:
#     #     batch["showershape"] = analyze_layers(batch)
#     # if "response" in metrics:
#     #     batch["response"] = response(batch)

#     assert torch.allclose(batch.x, batch_untf.x)
#     pcs_out = pc_to_voxel(batch)
#     input_images = torch.stack([e[1].reshape(*dims) for e in pcs], 0)
#     assert torch.allclose(pcs_out.float(), input_images)



"""
histcdf: Calculate an approximation to the cumulative distribution function using histograms.
hist_distance: Calculate a distance metric between two 1D samples using histcdf.
calc_hist_dist: Convenience function to run hist_distance over multiple dimensions and convert the result to a numpy array.
"""

from typing import Optional

import numpy as np
import torch


def histcdf(
    arr: torch.Tensor, bins: torch.Tensor, weight: Optional[torch.Tensor] = None
) -> torch.Tensor:
    """
    Calculate an approximation to the cumulative distribution function (CDF) using histograms.

    Args:
        arr (torch.Tensor): Input tensor containing data points.
        bins (torch.Tensor): Bins for histogram calculation.
        weight (Optional[torch.Tensor]): Weights for each data point (optional).

    Returns:
        torch.Tensor: Approximated cumulative distribution function values.
    """
    dev = arr.device
    if weight is not None:
        weight = weight.cpu()
    val = torch.histogram(arr.cpu(), bins=bins.cpu(), weight=weight)
    cdf = val.hist.cumsum(-1)
    cdf /= cdf[-1].clone()
    return cdf.to(dev)


def hist_distance(
    r: torch.Tensor,
    f: torch.Tensor,
    bins: torch.Tensor,
    rw: Optional[torch.Tensor] = None,
    fw: Optional[torch.Tensor] = None,
) -> torch.Tensor:
    """
    Calculate a distance metric between two 1D samples using the `histcdf` function.

    Args:
        r (torch.Tensor): Reference sample.
        f (torch.Tensor): Comparison sample.
        bins (torch.Tensor): Bins for histogram calculation.
        rw (Optional[torch.Tensor]): Weights for the reference sample (optional).
        fw (Optional[torch.Tensor]): Weights for the comparison sample (optional).

    Returns:
        torch.Tensor: Calculated distance metric between the two samples.
    """
    cdf_r = histcdf(r, bins, rw)
    cdf_f = histcdf(f, bins, fw)
    dist = (cdf_f - cdf_r).abs().mean()
    return dist


def calc_hist_dist(
    r: torch.Tensor,
    f: torch.Tensor,
    bins: torch.Tensor,
    rw: Optional[torch.Tensor] = None,
    fw: Optional[torch.Tensor] = None,
) -> np.ndarray:
    """
    Calculate distance metrics between two samples over multiple dimensions using `hist_distance` and convert to a numpy array.

    Args:
        r (torch.Tensor): Reference sample.
        f (torch.Tensor): Comparison sample.
        bins (torch.Tensor): Bins for histogram calculation.
        rw (Optional[torch.Tensor]): Weights for the reference sample (optional).
        fw (Optional[torch.Tensor]): Weights for the comparison sample (optional).

    Returns:
        np.ndarray: Array of distance metrics calculated for each dimension.
    """
    assert r.shape == f.shape
    assert (rw is None) == (fw is None)
    if len(r.shape) == 1:
        r = r.unsqueeze(-1)
        f = f.unsqueeze(-1)

    dists = []
    for iftx in range(r.shape[-1]):
        if isinstance(bins, list) and len(bins) == r.shape[1]:
            ibins = bins[iftx]
        elif isinstance(bins, torch.Tensor):
            ibins = bins
        else:
            raise Exception("Bins have the wrong shape/dimension.")
        idist = hist_distance(r[..., iftx], f[..., iftx], ibins, rw, fw)
        dists.append(idist)

    return torch.stack(dists, dim=0).cpu().numpy()



import torch
from torch_geometric.data import Batch
from torch_geometric.nn import global_add_pool, global_mean_pool
from torch_scatter import scatter_std


def sphereratio(batch: Batch) -> dict[str, torch.Tensor]:
    """
    Computes energy fractions within a sphere around the center of energy for each event in a batch,
    and returns the ratios of these energy fractions.

    The function calculates the energy fractions within two spheres (small and large) around the center
    of energy for each event in a batch. The centers are calculated as a weighted mean of the point
    coordinates, where the weights are the energies of the points.

    Parameters
    ----------
    batch : Batch
        A Batch object from the PyTorch Geometric library that contains the point cloud
        representation of the events.

    Returns
    -------
    dict
        A dictionary with keys 'small', 'large', and 'ratio', where:
            - 'small' corresponds to the tensor of energy fractions within the smaller sphere for each event.
            - 'large' corresponds to the tensor of energy fractions within the larger sphere for each event.
            - 'ratio' is the ratio of 'small' to 'large' for each event.
    """
    batchidx = batch.batch
    # Ehit = batch.x[:, conf.loader.x_ftx_energy_pos].reshape(-1, 1)
    Ehit = batch.x[:, 0].reshape(-1, 1)
    e_small, e_large = __dist_fraction(Ehit, batch.xyz, batchidx, 0.3, 0.8)

    return {
        "small": e_small,
        "large": e_large,
        "ratio": e_small / e_large,
    }


def cyratio(batch: Batch) -> dict[str, torch.Tensor]:
    """
    Similar to the sphereratio function, this function computes energy fractions within a cylinder around
    the center of energy for each event in a batch, and returns the ratios of these energy fractions.

    The function only considers the x and y coordinates of the points to calculate the energy fractions
    and the center of energy, effectively projecting the points onto the xy-plane and creating a cylindrical
    region.

    Parameters and return values are the same as those for the sphereratio function.
    """
    batchidx = batch.batch
    # Ehit = batch.x[:, conf.loader.x_ftx_energy_pos].reshape(-1, 1)
    Ehit = batch.x[:, 0].reshape(-1, 1)

    e_small, e_large = __dist_fraction(
        Ehit, batch.xyz[:, [0, 1]], batchidx, 0.2, 0.6
    )

    return {
        "small": e_small,
        "large": e_large,
        "ratio": e_small / e_large,
    }


def __dist_fraction(Ehit, pos, batchidx, small, large, center_energy_weighted=True):
    """
    Private helper function used by the sphereratio and cyratio functions to compute the energy fractions
    within certain regions around the center of energy.

    Parameters
    ----------
    Ehit : torch.Tensor
        Tensor of energies for each point in the batch.
    pos : torch.Tensor
        Tensor of positions for each point in the batch.
    batchidx : torch.Tensor
        Tensor of batch indices for each point in the batch.
    small : float
        Radius of the smaller region.
    large : float
        Radius of the larger region.
    center_energy_weighted : bool, optional
        Whether to calculate the center of energy as a weighted mean of the point coordinates.
        If False, the unweighted mean is used. Defaults to True.

    Returns
    -------
    tuple
        A tuple containing two tensors:
            - The energy fraction within the smaller region for each event.
            - The energy fraction within the larger region for each event.
    """
    Esum = global_add_pool(Ehit, batchidx).reshape(-1, 1)

    # get the center, weighted by energy
    if center_energy_weighted:
        center = global_add_pool(pos * Ehit, batchidx) / Esum
    else:
        center = global_mean_pool(pos, batchidx)
    std = scatter_std(pos, batchidx, dim=-2)
    # hit distance to center
    delta = (((pos - center[batchidx]) / std[batchidx]) ** 2).mean(-1).sqrt()
    del center, std
    # energy fraction inside circle around center
    e_small = (
        global_add_pool(Ehit.squeeze() * (delta < small).float(), batchidx)
        / Esum.squeeze()
    )
    e_large = (
        global_add_pool(Ehit.squeeze() * (delta < large).float(), batchidx)
        / Esum.squeeze()
    )
    # __plot_frac(e_small, e_large, small, large)
    return e_small, e_large



from typing import Optional

import torch
from torch_geometric.data import Batch
from torch_scatter import scatter_add

from .. import calorimeter
from .shift_multi_hits import shift_multi_hits
from .utils import fix_slice_dict_nodeattr, ptr_from_batchidx, scatter_sort


def shift_sum_multi_hits(batch, forbid_dublicates=False):
    """
    Faster combination of `shift_sum_multi_hits` and `shift_sum_multi_hits`.
    First shifts hits assigned to the same cell to empty neigboring cells, then
    sums the energy of then hits remaining in the same cell.
    If forbid_dublicates=False, the function also verifies that there were no duplicate hits in the original data.
    The function modifies the batch in place, updating batch.x, batch.batch, and batch.ptr.

    Parameters
    ----------
    batch : Batch
        A Batch object from the PyTorch Geometric library that contains the point cloud
        representation of the events. Batch.x contains the hit energy and 3D coordinates of hits.
        Batch.batch contains the indices that map which hit belongs to which shower.

    forbid_dublicates : bool, optional
        If True, asserts that there were no duplicate hits in the original data. Defaults to True.

    shiftmultihit : bool, optional
        If True, tries to move hits from overfilled cells to neighboring empty cells. Defaults to True.

    Returns
    -------
    batch : Batch
        The modified Batch object where duplicate hits have been summed.
    """

    # batch = batch.to("cpu")
    batchidx = batch.batch
    assert (batchidx.diff() >= 0).all()
    globalidx = calorimeter.globalidx_from_pos(batch.x[:, 1:].long(), batchidx)

    # get new positions and global index
    # and the current index of these events
    batch, globalidx = shift_multi_hits(
        batch, globalidx.clone(), return_globalidx=True
    )

    batch = sum_multi_hits(batch, globalidx, forbid_dublicates)
    return batch


def sum_multi_hits(
    batch: Batch,
    globalidx: Optional[torch.Tensor] = None,
    forbid_dublicates: bool = False,
):
    """
    Sums the energy of duplicate hits in the same cell for each event.
    If fake=False, the function also verifies that there were no duplicate hits in the original data.
    The function modifies the batch in place, updating batch.x, batch.batch, and batch.ptr.

    Parameters
    ----------
    batch : Batch
        A Batch object from the PyTorch Geometric library that contains the point cloud
        representation of the events. Batch.x contains the hit energy and 3D coordinates of hits.
        Batch.batch contains the indices that map which hit belongs to which shower.

    forbid_dublicates : bool, optional
        If True, asserts that there were no duplicate hits in the original data. Defaults to True.

    Returns
    -------
    batch : Batch
        The modified Batch object where duplicate hits have been summed.
    """
    dev = batch.x.device
    batchidx = batch.batch

    if globalidx is None:
        globalidx: torch.Tensor = calorimeter.globalidx_from_pos(
            batch.x[:, 1:].long(), batchidx
        )

    # sort the globalidx
    globalidx, index_perm = scatter_sort(globalidx, batchidx)
    batch.x = batch.x[index_perm]

    hitE = batch.x[:, 0]
    pos = batch.x[:, 1:].long()
    assert (batchidx[index_perm] == batchidx).all()

    # unique_cells_idx counts up every time a new cell in an
    # even is accessed in globalidx
    # counts gives the times the cell/event idx is occupied
    _, unique_cells_idx, counts = torch.unique(
        globalidx, return_inverse=True, return_counts=True
    )

    if forbid_dublicates:
        assert (counts - 1 == 0).all()

    # begin sum
    hitE_new = scatter_add(hitE, unique_cells_idx)
    sel_new_idx = counts.cumsum(-1) - 1
    if forbid_dublicates:
        assert (sel_new_idx == torch.arange(len(batch.x)).to(dev)).all()

    batchidx_new = batchidx[sel_new_idx]
    pos_new = pos[sel_new_idx]

    # count the cells, that have been hit multiple times
    n_multihit = scatter_add(counts - 1, batchidx_new)
    if forbid_dublicates:
        assert (n_multihit == 0).all()
    new_counts = torch.unique_consecutive(batchidx_new, return_counts=True)[1]

    x_new = torch.hstack([hitE_new.reshape(-1, 1), pos_new])

    # # TODO remove sanity test:
    # old_counts = torch.unique_consecutive(batchidx, return_counts=True)[1]
    # if "n_pointsv" in batch.keys:
    #     assert (old_counts == batch.n_pointsv).all()
    # assert ((old_counts - new_counts) == n_multihit).all()
    # assert torch.allclose(
    #     scatter_add(hitE_new, batchidx_new), scatter_add(hitE, batchidx)
    # )

    if forbid_dublicates:
        assert (n_multihit == 0).all()
        assert (batch.batch == batchidx_new).all()
        assert (batch.n_pointsv == new_counts).all()
        for i in range(4):
            assert torch.allclose(batch.x.T[i][index_perm], x_new.T[i])

    batch.n_multihit = n_multihit
    batch.batch = batchidx_new
    batch.x = x_new
    batch.n_pointsv = new_counts
    # need to shift the ptr by the number of removed hits
    batch.ptr = ptr_from_batchidx(batchidx_new)

    batch.nhits = {
        "n": batch.n_pointsv,
        "n_by_E": batch.n_pointsv / batch.y[:, 0],
    }
    fix_slice_dict_nodeattr(batch, "x")

    return batch.to(dev)



from math import prod
from typing import Optional

import torch


class Calorimeter:
    def __init__(self) -> None:
        self._dims: Optional[tuple[int, int, int]] = None
        self._num_r: Optional[int] = None
        self._num_alpha: Optional[int] = None
        self._num_z: Optional[int] = None
        self._caloname: Optional[str] = None

    def _assert_calo_init(self):
        if self._caloname is None:
            raise Exception(
                "Calorimeter is not initalized. Use"
                " `init_calorimeter(caloname)` before accassing the attributes."
            )

    @property
    def cell_idxs(self) -> torch.Tensor:
        return torch.arange(prod(self.dims)).reshape(*(self.dims))

    @property
    def num_r(self) -> int:
        self._assert_calo_init()
        return self._num_r

    @property
    def num_alpha(self) -> int:
        self._assert_calo_init()
        return self._num_alpha

    @property
    def num_z(self) -> int:
        self._assert_calo_init()
        return self._num_z

    @property
    def dims(self) -> tuple[int, int, int]:
        self._assert_calo_init()
        return self._dims

    def pos_to_cellidx(self, pos: torch.Tensor) -> torch.Tensor:
        """Map the position of a point to index unique for every cell:

        Parameters
        ----------
        pos : Tensor
            The position matrix of the points with shape `[n_points,n_dims]`.

        """
        assert pos.shape[-1] == len(self.dims)
        dev = pos.device
        pos = pos.long()
        return self.cell_idxs.to(dev)[pos[..., 0], pos[..., 1], pos[..., 2]]

    def globalidx_from_pos(
        self, pos: torch.Tensor, batchidx: torch.Tensor
    ) -> torch.Tensor:
        """Map the position of a point to index unique for every cell and every event:

        Parameters
        ----------
        pos : Tensor
            The position matrix of the points with shape `[n_points,n_dims]`.
        batchidx : Tensor
            The tensor indicating the event number of each point with shape `[n_points]`.

        """
        num_z, num_alpha, num_r = self.dims
        pos_z, pos_alpha, pos_r = pos.T
        assert (pos >= 0).all()
        assert (pos_z < num_z).all()
        assert (pos_alpha < num_alpha).all()
        assert (pos_r < num_r).all()
        n_events = int(batchidx[-1] + 1)
        # create and index that indexes each cell for each event
        globalidx = self.pos_to_cellidx(pos)
        # add the shiftvector to diffentiate for each event
        eventshift = torch.arange(n_events).to(pos.device) * prod(calorimeter.dims)
        globalidx = globalidx + eventshift[batchidx]
        return globalidx

    def init_calorimeter(self, caloname: str):
        """The function `init_calorimeter` initializes a calorimeter geometry.
        Currently implemented are dataset 2 and 3 of the CaloChallenge:
        https://github.com/CaloChallenge/homepage

        Parameters
        ----------
        caloname : str
            The parameter `caloname` is a string that represents the name of the calorimeter. It can have two
        possible values: "cc_ds2" or "cc_ds3".

        """
        match caloname:
            case "cc_ds2":
                layout = (45, 16, 9)
            case "cc_ds3":
                layout = (45, 50, 18)
            case "test":
                layout = (2, 2, 1)
            case "test_large":
                layout = (4, 3, 4)
            case _:
                raise NotImplementedError(
                    f"No such calorimeter: {caloname}. Options are"
                    " :'cc_ds2','cc_ds3'"
                )
        self._caloname = caloname
        self._num_z, self._num_alpha, self._num_r = layout
        self._dims = layout


calorimeter = Calorimeter()



from collections import defaultdict

import torch
import torch_scatter
from torch import LongTensor, Tensor
from torch_geometric.data import Batch


def init_batch(batch_idx: LongTensor):
    if not batch_idx.dtype == torch.long:
        raise Exception("Batch index dtype must be torch.long")
    if not (batch_idx.diff() >= 0).all():
        raise Exception("Batch index must be increasing")
    if not batch_idx.dim() == 1:
        raise Exception()

    batch = Batch(batch=batch_idx)

    batch.ptr = ptr_from_batchidx(batch_idx)
    batch._num_graphs = int(batch.batch.max() + 1)

    batch._slice_dict = defaultdict(dict)
    batch._inc_dict = defaultdict(dict)

    return batch


def ptr_from_batchidx(batch_idx: LongTensor):
    # Construct the ptr to adress single graphs
    assert batch_idx.dtype == torch.long
    # graph[idx].x== batch.x[batch.ptr[idx]:batch.ptr[idx]+1]
    # Get delta with diff
    # Get idx of diff >0 with nonzero
    # shift by -1
    # add the batch size -1 as last element and add 0 in front
    dev = batch_idx.device
    return torch.concatenate(
        (
            torch.tensor(0).long().to(dev).unsqueeze(0),
            (batch_idx.diff()).nonzero().reshape(-1) + 1,
            torch.tensor(len(batch_idx)).long().to(dev).unsqueeze(0),
        )
    )


def add_node_attr(batch: Batch, attrname: str, attr: Tensor):
    device = attr.device
    assert device == batch.batch.device
    batch_idxs = batch.batch

    batch[attrname] = attr
    out = torch_scatter.scatter_add(
        torch.ones(len(attr), dtype=torch.long, device=device), batch_idxs, dim=0
    )
    out = out.cumsum(dim=0)
    # batch._slice_dict[attrname] =  torch.cat(
    #     [torch.zeros(1, dtype=torch.long, device=device), out], dim=0
    # )
    batch._slice_dict[attrname] = pad_zero(out).cpu()

    batch._inc_dict[attrname] = torch.zeros(
        batch._num_graphs, dtype=torch.long, device=device
    ).cpu()


def add_graph_attr(batch: Batch, attrname: str, attr: Tensor):
    device = attr.device
    assert device == batch.batch.device

    batch[attrname] = attr
    batch._slice_dict[attrname] = torch.arange(
        batch.num_graphs + 1, dtype=torch.long, device=device
    ).cpu()

    batch._inc_dict[attrname] = torch.zeros(
        batch.num_graphs, dtype=torch.long, device=device
    ).cpu()


def set_edge_attr(
    batch: Batch, edge_attr: LongTensor, batchidx_per_edge: LongTensor
):
    assert batchidx_per_edge.dtype == torch.long
    assert hasattr(batch, "edge_index") and batch["edge_index"].dtype == torch.long
    batch.edge_attr = edge_attr
    # Fix _slice_dict
    # edges_per_graph = batchidx_per_edge.unique(return_counts=True)[1]
    # batch._slice_dict["edge_attr"] = pad_zero(edges_per_graph.cumsum(0))
    batch._slice_dict["edge_attr"] = batch._slice_dict["edge_index"].cpu()
    batch._inc_dict["edge_attr"] = torch.zeros(
        batch.num_graphs, device=edge_attr.device
    ).cpu()


def set_edges(batch: Batch, edges: LongTensor, batchidx_per_edge: LongTensor):
    assert edges.dtype == batchidx_per_edge.dtype == torch.long
    assert edges.device == batchidx_per_edge.device == batch.batch.device
    assert (batchidx_per_edge.diff() >= 0).all(), "Edges must be ordered by batch"
    if batch.edge_index is None:
        batch.edge_index = torch.empty(
            2, 0, dtype=torch.long, device=batch.batch.device
        )
    # Edges must be shifted by the number sum of the nodes in the previous graphs
    edges += batch.ptr[batchidx_per_edge]
    batch.edge_index = torch.hstack((batch.edge_index.clone(), edges))
    # Fix _slice_dict
    edges_per_graph = batchidx_per_edge.unique(return_counts=True)[1]
    batch._slice_dict["edge_index"] = pad_zero(edges_per_graph.cumsum(0)).cpu()
    batch._inc_dict["edge_index"] = batch.ptr[:-1].cpu()


def pad_zero(arr: torch.Tensor):
    return torch.cat(
        [torch.tensor(0, dtype=arr.dtype, device=arr.device).unsqueeze(0), arr]
    )


def pad_one(arr: torch.Tensor):
    return torch.cat(
        [torch.tensor(1, dtype=arr.dtype, device=arr.device).unsqueeze(0), arr]
    )


def from_batch_list(*batches: list[Batch]):
    res = Batch.from_data_list(batches)
    res.ptr = ptr_from_batchidx(res.batch)
    k = "edge_attr"
    for k in res.keys:
        res._slice_dict[k] = torch.cat(
            [
                be._slice_dict[k]
                + (0 if ibatch == 0 else batches[ibatch - 1]._slice_dict[k][-1])
                for ibatch, be in enumerate(batches)
            ]
        )
        # res._inc_dict[k] = torch.cat(
        #     [
        #         be._slice_dict[k]
        #         for ibatch, be in enumerate(batches)
        #     ]
        # )
    return res



from math import prod

import torch
from torch_geometric.data import Batch
from torch_scatter import scatter_add

from .. import calorimeter
from .utils import _construct_global_cellevent_idx


def pc_to_voxel(batch: Batch) -> torch.Tensor:
    """
    Converts a pytorch geometric batch of point clouds into a torch batch of 3D voxel grids.

    Parameters
    ----------
    batch : Batch
        A Batch object from the PyTorch Geometric library that contains the point cloud
        representation of the events.

    Returns
    -------
    torch.Tensor
        A tensor of shape (batch_size, num_z, num_alpha, num_r), where each element
        represents the energy in the corresponding voxel.
    """
    dims = calorimeter.dims
    batch_size = int(batch.batch[-1] + 1)
    x = batch.x
    shower_index = batch.batch
    Ehit = x.T[0]
    valid_coordinates = x.T[1:].int()
    indices = torch.cat((shower_index.unsqueeze(1), valid_coordinates.t()), dim=1)

    full_event_cell_idx = _construct_global_cellevent_idx(batch_size).to(x.device)
    scatter_index = full_event_cell_idx[
        indices[..., 0], indices[..., 1], indices[..., 2], indices[..., 3]
    ]
    vox = scatter_add(
        src=Ehit,
        index=scatter_index,
        dim_size=prod(dims) * batch_size,
    )
    return vox.reshape(batch_size, *dims)
